#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "vm/Operation.h"
#include "vm/Script.h"

static void sError(Script* sc, const char* format, ...) {
    va_list args;
    va_start(args, format);
    vsnprintf(sc->error, SCRIPT_ERROR_SIZE, format, args);
    va_end(args);
}

static void sIntPrinter(int i) {
    printf("%d\n", i);
}

static void sFloatPrinter(float f) {
    printf("%.2f\n", f);
}

static void sBoolPrinter(bool b) {
    puts(b ? "true" : "false");
}

static IntPrinter intPrinter = sIntPrinter;
static FloatPrinter floatPrinter = sFloatPrinter;
static BoolPrinter boolPrinter = sBoolPrinter;

static bool sRead(Script* sc, void* buffer, int length) {
    if(sc->readIndex + length > sc->code->length) {
        sError(sc, "cannot read expected %d bytes of data from bytecode");
        return false;
    }
    memcpy(buffer, sc->code->code + sc->readIndex, length);
    sc->readIndex += length;
    return true;
}

static Operation sReadOperation(Script* sc) {
    unsigned char c;
    if(sRead(sc, &c, 1)) {
        return c;
    }
    return OP_NOTHING;
}

static void* sReserve(Script* sc, int length) {
    if(sc->stackIndex + length > SCRIPT_STACK_SIZE) {
        sError(sc, "stack overflow");
        return NULL;
    }
    void* p = sc->stack + sc->stackIndex;
    sc->stackIndex += length;
    return p;
}

static bool sPush(Script* sc, const void* data, int length) {
    void* p = sReserve(sc, length);
    if(p != NULL) {
        memcpy(p, data, length);
        return true;
    }
    return false;
}

static const void* sFree(Script* sc, int length) {
    if(sc->stackIndex < length) {
        sError(sc, "stack underflow");
        return NULL;
    }
    sc->stackIndex -= length;
    return sc->stack + sc->stackIndex;
}

static bool sPop(Script* sc, void* data, int length) {
    const void* p = sFree(sc, length);
    if(p != NULL) {
        memcpy(data, p, length);
        return true;
    }
    return false;
}

static bool sPeek(Script* sc, void* data, int length) {
    if(sc->stackIndex < length) {
        sError(sc, "stack underflow");
        return false;
    }
    memcpy(data, sc->stack + (sc->stackIndex - length), length);
    return true;
}

#define POP_PUSH(type, Type)                                                   \
    static bool sPop##Type(Script* sc, type* value) {                          \
        return sPop(sc, value, sizeof(type));                                  \
    }                                                                          \
    static bool sPush##Type(Script* sc, type value) {                          \
        return sPush(sc, &value, sizeof(type));                                \
    }

#define READ_POP_PUSH(type, Type)                                              \
    static bool sRead##Type(Script* sc, type* i) {                             \
        return sRead(sc, i, sizeof(type));                                     \
    }                                                                          \
    POP_PUSH(type, Type)

#define PUSH_CONSTANT(type, Type, init)                                        \
    {                                                                          \
        type value = init;                                                     \
        if(sRead##Type(sc, &value)) {                                          \
            sPush##Type(sc, value);                                            \
        }                                                                      \
    }

#define ZERO_CHECK(name)                                                       \
    if(values[0] == 0) {                                                       \
        sError(sc, name " by 0");                                              \
    }

#define OP_BASE(type, Type, RType, op, check)                                  \
    {                                                                          \
        type values[2];                                                        \
        if(sPop##Type(sc, values) && sPop##Type(sc, values + 1)) {             \
            check;                                                             \
            sPush##RType(sc, values[1] op values[0]);                          \
        }                                                                      \
    }

#define CHECKED_NUMBER_OP(type, Type, op, check)                               \
    OP_BASE(type, Type, Type, op, check)
#define NUMBER_OP(type, Type, op) CHECKED_NUMBER_OP(type, Type, op, )
#define BOOL_OP(type, Type, op) OP_BASE(type, Type, Bool, op, )

READ_POP_PUSH(int, Int)
READ_POP_PUSH(float, Float)
POP_PUSH(bool, Bool)

#define PRINT(type, Type, init)                                                \
    {                                                                          \
        type value = init;                                                     \
        if(sPop##Type(sc, &value)) {                                           \
            type##Printer(value);                                              \
        }                                                                      \
    }

static void sReserveBytes(Script* sc) {
    int bytes = 0;
    int offset = 0;
    if(sReadInt(sc, &bytes) && sReadInt(sc, &offset)) {
        int oldIndex = sc->stackVarIndex;
        sc->stackVarIndex = sc->stackIndex - offset;
        sReserve(sc, bytes - offset);
        sPushInt(sc, oldIndex);
    }
}

static bool sCheckAddress(Script* sc, int address, int length) {
    if(address < 0 || address + length > sc->stackIndex) {
        sError(sc, "address is out of stack bounds");
        return false;
    }
    return true;
}

static void sLoad(Script* sc, int length) {
    int address = -1;
    if(sReadInt(sc, &address)) {
        address += sc->stackVarIndex;
        if(sCheckAddress(sc, address, length)) {
            sPush(sc, sc->stack + address, length);
        }
    }
}

static void sStore(Script* sc, int length) {
    int address = -1;
    if(sReadInt(sc, &address)) {
        address += sc->stackVarIndex;
        if(sCheckAddress(sc, address, length)) {
            sPop(sc, sc->stack + address, length);
        }
    }
}

/*static void sAllocateArray(Script* sc) {
    Object length;
    if(!sPop(sc, &length) && sCheckType(sc, &length, OT_INT)) {
        if(length.as.intValue < 0) {
            sError(sc, "negative array length");
            return;
        }
        Object o = {.type = OT_ARRAY,
                    .as.intValue =
                        aAllocate(&sc->allocator, length.as.intValue)};
        sPush(sc, &o);
    }
}

static void sArrayLength(Script* sc) {
    Object* o = sPopAndDereference(sc);
    if(o != NULL && sCheckType(sc, o, OT_ARRAY)) {
        int arrayPointer = o->as.intValue;
        if(arrayPointer < 0 || arrayPointer >= sc->allocator.capacity) {
            sError(sc, "array pointer is out of range");
            return;
        }
        sPushInt(sc, sc->allocator.data[arrayPointer].length);
    }
}*/

static void sPushCodeString(Script* sc) {
    int value = 0;
    if(sReadInt(sc, &value)) {
        char* s = (char*)(sc->code->code + sc->readIndex);
        sc->readIndex += value;
        sPush(sc, s, sizeof(char*));
    }
}

static void sInvertSignInt(Script* sc) {
    int i = 0;
    if(sPopInt(sc, &i)) {
        sPushInt(sc, -i);
    }
}

static void sInvertSignFloat(Script* sc) {
    float f = 0;
    if(sPopFloat(sc, &f)) {
        sPushFloat(sc, -f);
    }
}

static void sNot(Script* sc) {
    bool value = false;
    if(sPopBool(sc, &value)) {
        sPushBool(sc, !value);
    }
}

static void sBitNot(Script* sc) {
    int value = 0;
    if(sPopInt(sc, &value)) {
        sPushInt(sc, ~value);
    }
}

static void sLine(Script* sc) {
    sRead(sc, &sc->line, 2);
}

static void sGoTo(Script* sc) {
    int gotoIndex;
    if(sReadInt(sc, &gotoIndex)) {
        sc->readIndex = gotoIndex;
    }
}

static void sGoSub(Script* sc) {
    int gotoIndex;
    int offset;
    if(sReadInt(sc, &gotoIndex) && sReadInt(sc, &offset)) {
        int address = sc->stackIndex - offset - sizeof(int);
        if(sCheckAddress(sc, address, sizeof(int))) {
            memcpy(sc->stack + address, &sc->readIndex, sizeof(int));
            sc->readIndex = gotoIndex;
        }
    }
}

static void sReturn(Script* sc) {
    int bytes = 0;
    int varIndex = 0;
    if(sReadInt(sc, &bytes) && sPopInt(sc, &varIndex)) {
        sc->stackVarIndex = varIndex;
        sFree(sc, bytes);

        int returnIndex;
        if(sPopInt(sc, &returnIndex)) {
            sc->readIndex = returnIndex;
        }
    }
}

static void sReturnInt(Script* sc) {
    int value = 0;
    if(sPopInt(sc, &value)) {
        sReturn(sc);
        sPushInt(sc, value);
    }
}

static void sReturnBool(Script* sc) {
    bool value = 0;
    if(sPopBool(sc, &value)) {
        sReturn(sc);
        sPushBool(sc, value);
    }
}

static void sIfGoTo(Script* sc) {
    int gotoIndex = 0;
    bool value = false;
    if(sReadInt(sc, &gotoIndex) && sPopBool(sc, &value) && !value) {
        sc->readIndex = gotoIndex;
    }
}

static void sPeekFalseGoTo(Script* sc) {
    int gotoIndex = 0;
    bool value = false;
    if(sReadInt(sc, &gotoIndex) && sPeek(sc, &value, sizeof(bool)) && !value) {
        sc->readIndex = gotoIndex;
    }
}

static void sPeekTrueGoTo(Script* sc) {
    int gotoIndex = 0;
    bool value = false;
    if(sReadInt(sc, &gotoIndex) && sPeek(sc, &value, sizeof(bool)) && value) {
        sc->readIndex = gotoIndex;
    }
}

static void sConsumeInstruction(Script* sc) {
    switch(sReadOperation(sc)) {
        case OP_NOTHING: break;
        case OP_PUSH_INT: PUSH_CONSTANT(int, Int, 0); break;
        case OP_PUSH_FLOAT: PUSH_CONSTANT(float, Float, 0.0f); break;
        case OP_PUSH_CONST_STRING: sPushCodeString(sc); break;
        case OP_PUSH_TRUE: sPushBool(sc, true); break;
        case OP_PUSH_FALSE: sPushBool(sc, false); break;
        case OP_ADD_INT: NUMBER_OP(int, Int, +); break;
        case OP_ADD_FLOAT: NUMBER_OP(float, Float, +); break;
        case OP_SUB_INT: NUMBER_OP(int, Int, -); break;
        case OP_SUB_FLOAT: NUMBER_OP(float, Float, -); break;
        case OP_MUL_INT: NUMBER_OP(int, Int, *); break;
        case OP_MUL_FLOAT: NUMBER_OP(float, Float, *); break;
        case OP_DIV_INT:
            CHECKED_NUMBER_OP(int, Int, /, ZERO_CHECK("division"));
            break;
        case OP_DIV_FLOAT:
            CHECKED_NUMBER_OP(float, Float, /, ZERO_CHECK("division"));
            break;
        case OP_MOD_INT:
            CHECKED_NUMBER_OP(int, Int, %, ZERO_CHECK("module"));
            break;
        case OP_INVERT_SIGN_INT: sInvertSignInt(sc); break;
        case OP_INVERT_SIGN_FLOAT: sInvertSignFloat(sc); break;
        case OP_LESS_INT: BOOL_OP(int, Int, <); break;
        case OP_LESS_FLOAT: BOOL_OP(float, Float, <); break;
        case OP_GREATER_INT: BOOL_OP(int, Int, >); break;
        case OP_GREATER_FLOAT: BOOL_OP(float, Float, >); break;
        case OP_EQUAL_INT: BOOL_OP(int, Int, ==); break;
        case OP_EQUAL_FLOAT: BOOL_OP(float, Float, ==); break;
        case OP_EQUAL_BOOL: BOOL_OP(bool, Bool, ==); break;
        case OP_NOT: sNot(sc); break;
        case OP_AND: BOOL_OP(bool, Bool, &&); break;
        case OP_OR: BOOL_OP(bool, Bool, ||); break;
        case OP_BIT_NOT: sBitNot(sc); break;
        case OP_BIT_AND: NUMBER_OP(int, Int, &); break;
        case OP_BIT_OR: NUMBER_OP(int, Int, |); break;
        case OP_BIT_XOR: NUMBER_OP(int, Int, ^); break;
        case OP_LEFT_SHIFT: NUMBER_OP(int, Int, <<); break;
        case OP_RIGHT_SHIFT: NUMBER_OP(int, Int, >>); break;
        case OP_PRINT_INT: PRINT(int, Int, 0); break;
        case OP_PRINT_FLOAT: PRINT(float, Float, 0.0f); break;
        case OP_PRINT_BOOL: PRINT(bool, Bool, false); break;
        case OP_LINE: sLine(sc); break;
        case OP_GOTO: sGoTo(sc); break;
        case OP_IF_GOTO: sIfGoTo(sc); break;
        case OP_PEEK_FALSE_GOTO: sPeekFalseGoTo(sc); break;
        case OP_PEEK_TRUE_GOTO: sPeekTrueGoTo(sc); break;
        case OP_GOSUB: sGoSub(sc); break;
        case OP_RETURN: sReturn(sc); break;
        case OP_RETURN_INT: sReturnInt(sc); break;
        case OP_RETURN_BOOL:
            sReturnBool(sc);
            break;
            // case OP_SET_RETURN: sSetReturn(sc); break;
            // case OP_DUPLICATE: sDuplicate(sc); break;
            // case OP_ALLOCATE_ARRAY: sAllocateArray(sc); break;
            // case OP_ARRAY_LENGTH: sArrayLength(sc); break;
        case OP_RESERVE: sReserveBytes(sc); break;
        case OP_LOAD_INT: sLoad(sc, sizeof(int)); break;
        case OP_STORE_INT: sStore(sc, sizeof(int)); break;
    }
    // sCollectGarbage(sc);
}

static bool sHasData(Script* sc) {
    return sc->readIndex < sc->code->length;
}

Script* sInit(ByteCode* code) {
    Script* sc = malloc(sizeof(Script));
    sc->error[0] = '\0';
    sc->code = code;
    sc->readIndex = 0;
    sc->stackIndex = 0;
    sc->stackVarIndex = 0;
    sc->line = 0;
    // aInit(&sc->allocator);
    return sc;
}

void sDelete(Script* sc) {
    bcDelete(sc->code);
    // aDelete(&sc->allocator);
    free(sc);
}

void sRun(Script* sc) {
    while(sHasData(sc)) {
        sConsumeInstruction(sc);
        if(sc->error[0] != '\0') {
            puts("error:");
            printf(" - info: %s\n", sc->error);
            printf(" - line: %d\n", sc->line);
            return;
        }
    }
}

void sSetIntPrinter(IntPrinter p) {
    intPrinter = p;
}

void sSetFloatPrinter(FloatPrinter p) {
    floatPrinter = p;
}

void sSetBoolPrinter(BoolPrinter p) {
    boolPrinter = p;
}

/*static void sMark(Script* sc, Object* o) {
    if(o->type == OT_ARRAY) {
        Array* a = sc->allocator.data + o->as.intValue;
        a->marked = true;
        for(int i = 0; i < a->length; i++) {
            sMark(sc, a->data + i);
        }
    }
}

void sCollectGarbage(Script* sc) {
    aClearMarker(&sc->allocator);
    for(int i = 0; i < sc->stackIndex; i++) {
        sMark(sc, sc->stack + i);
    }
    aRemoveUnmarked(&sc->allocator);
}*/